# !./bin/python3
# *-* coding: utf-8 *-*
'''
SnippyCopy
=============================================================================
:: Description
    short description
'''

from __future__ import annotations

import ctypes

def clipboard() -> str:
    '''

    '''    
    lib  : ctypes.LibraryLoader = ctypes.windll
    void : ctypes.c_void_p = ctypes.c_void_p
    
    kernel_lock   = lib.kernel32.GlobalLock
    kernel_unlock = lib.kernel32.GlobalUnlock

    kernel_lock.argtypes = [void]
    kernel_lock.restype = void
    kernel_unlock.argtypes = [void]

    user = lib.user32
    clipboardData = user.GetClipboardData
    clipboardData.restype = void
    user.OpenClipboard(0)
    
    try:
        if user.IsClipboardFormatAvailable(1):
            data = clipboardData(1)
            data_locked = kernel_lock(data)
            text = ctypes.c_char_p(data_locked)
            value = text.value
            kernel_unlock(data_locked)
            return value.decode('utf-8')
    finally:
            user.CloseClipboard()

if __name__ == '__main__':
    pass


# !./bin/python3
# *-* coding: utf-8 *-*
'''
SnippyJson
=============================================================================
:: Description
    short description
'''
from __future__ import annotations

from dataclasses import dataclass

from .snippyCopy import clipboard
from .snippyDocument import Document

@dataclass
class Json:
    '''
    '''
    prefix: str
    body: str
    description: str

    def __repr__(self) -> str:
        return str(self.content)
        
    @property
    def name(self):
        return f'Snippycode.{self.prefix}'

    @property
    def content(self):
        return {self.name:{'prefix' :self.prefix,
                           'body'   :self.body,
                           'description':self.description}}


@dataclass
class JsonFromClipboard(Json):
    '''
    '''
    body : str = clipboard()
    description: str = '# Snippy from `clipboard`.'

    def __post_init__(self):
        super().__init_subclass__()


@dataclass
class JsonFromDocument(Json):
    '''
    '''
    path : str

    def __post_init__(self):
        document = Document(self.path)
        self.body : str = document.body
        self.description: str = document.description
        super().__init_subclass__()

if __name__ == '__main__':
    pass


# !./bin/python3
# *-* coding: utf-8 *-*
'''
SnippyFile
=============================================================================
:: Description
    short description
'''

from __future__ import annotations

from json import loads, dumps
from os import getenv
from os.path import join, exists

from .snippyJson import Json, JsonFromDocument, JsonFromClipboard

ENCODING: str = "utf-8"
HELLO   : dict = {"Snippycode":{
                        "prefix":"Hello!",
                        "body":"# Thank you for downloading SnippyCode!",
                        "description":"Welcome to SnippyCode!" }}
PATH = join(getenv('appdata'), 'Code/User/snippets/python.json' )


class File:
    '''
    '''
    def __init__(self) -> None:
        self.path = PATH
        if not(exists(self.path) 
               and self.read()):
            self.write(HELLO)

    def __repr__(self) -> str:
        return str(self.keys)

    @property
    def keys(self) -> list:
        return [*self.read().keys()]

    def read(self) -> dict[dict[str]]:
        with open(self.path, mode='r',
                  encoding=ENCODING) as file:
            content = file.read()
        return loads(content) if content else dict()
        
    def update(self, name, body) -> None:
        file: dict = self.read()
        file[name] |= {'body':body}
        self.write(file)

    def write(self, content= {}) -> None:
        old_content = self.read()
        with open(self.path, mode='w',
                  encoding=ENCODING) as file:
            file.write(dumps(old_content |
                             content,
                             indent=4))

    def delete(self, name) -> None:
        file: dict = self.read()
        del file[name]
        self.write(file)


class SnippetFile(File):
    '''
    '''
    def __init__(self, json: Json) -> None:
        self.json = json
        super().__init__()

    def __repr__(self) -> str:
        return super().__repr__()

    def update(self) -> None:
        return super().update(self.json.name, self.json.body)

    def write(self) -> None:
        return super().write(self.json.content)

    def delete(self) -> None:
        return super().delete(self.json.name)


class SnippetFromDocument(SnippetFile):
    '''
    '''
    def __init__(self, prefix: str, path: str) -> None:
        json = JsonFromDocument(prefix, None, None, path)
        super().__init__(json)


class SnippetFromClipboard(SnippetFile):
    '''
    '''
    def __init__(self, prefix: str) -> None:
        json = JsonFromClipboard(prefix)
        super().__init__(json)


if __name__ == '__main__':
    pass


# !./bin/python3
# *-* coding: utf-8 *-*
'''
SnippyStatus
=============================================================================
:: Description
    short description
'''
from __future__ import annotations

from dataclasses import dataclass

@dataclass
class Status:
    '''
    Print a status message with different formats.
    '''
    message: str

    def fmt_print(self, format:str):
        message = '╍╍╍✂ ' + \
        '╌'*72+'\n' + \
        f'{self.message:^78s}'

        print(f'\033{format}{message}\033[0m')

    @property
    def bold(self):
        self.fmt_print('[1m')
        
    @property
    def success(self):
        self.fmt_print('[92m')

    @property
    def warning(self):
         self.fmt_print('[93m')

    @property
    def failed(self):
        self.fmt_print('[91m')

if __name__ == '__main__':
    pass


# !./bin/python3
# *-* coding: utf-8 *-*
'''
SnippyCode

:: Description
    A command line extension to add snippets directly in Visual Studio Code.

:: Copyright
    SnippyCode, Copyright (C) 2022, Fabio Craig Wimmer Florey 
    All Rights Reserved.
'''
from __future__ import annotations

from argparse import ArgumentParser

from .snippyFile import SnippetFromDocument, SnippetFromClipboard
from .snippyStatus import Status


parser = ArgumentParser(description = __doc__)

parser.add_argument('name', 
                    help  = 'Name that you will type into vscode.',
                    nargs = '?')
parser.add_argument('path', 
                    help  = 'Path of the file you want to snip.',
                    nargs = '?') 

arguments = parser.parse_args()

json = (SnippetFromDocument(arguments.name, arguments.path) 
            if arguments.path else
            SnippetFromClipboard(arguments.name))

def list():
    Status(f'You have the following snippets:').bold
    for snippet in json.keys:
        print(snippet)

def write():
    try:
        json.write()
        Status(f'Successfully added `{json.json.prefix}` to your snippets!').success
    except Exception:
        Status(Exception).failed

def update():
    try:
        json.update()
        Status(f'Successfully updated `{json.json.prefix}` in your snippets!').success
    except Exception:
        Status(Exception).failed

def delete():
    try:
        json.delete()
        Status(f'Successfully deleted `{json.json.prefix}` from your snippets!').success
    except Exception:
        Status(Exception).failed

if __name__ == '__main__':
    pass


# !./bin/python3
# *-* coding: utf-8 *-*
'''
SnippyDocument
=============================================================================
:: Description
    short description
'''
from __future__ import annotations

from ast import get_docstring, parse
from dataclasses import dataclass
from os.path import basename

@dataclass
class Document:
    path : str

    def __post_init__(self):
        self.read()

    def read(self) -> str:
        with open(self.path, mode='r',
                    encoding='utf-8') as file:
            self.body = file.read()

    @property
    def file(self) -> str:
        return basename(self.path)

    @property
    def description(self) -> str:
        description = get_docstring(parse(self.body))
        return (description 
                if description
                else f"# Snippy from `{self.file}`.")

if __name__ == '__main__':
    pass


# !./bin/python3
# *-* coding: utf-8 *-*
'''
                                     SETUP
===============================================================================
                               Short Description.
-------------------------------------------------------------------------------

:Description:
   ...

:Usage:
    ...

:Authors:
    Fabio Craig Wimmer Florey

:Version:
    0.0.1-dev

'''
from __future__ import annotations

from dataclasses import dataclass
from os.path import abspath, dirname, exists, join
from pkg_resources import get_distribution, DistributionNotFound
from subprocess import check_call
from sys import executable

from setuptools import find_packages, setup

ENCODING = ("encoding", "utf8")

# NOTE :: Custom Classes and Functions.

@dataclass
class Status:
    '''
    Print a status message with different formats.
    '''
    message: str

    def fmt_print(self, format:str):
        print(f'\033{format}{self.message}\033[0m')

    @property
    def bold(self):
        self.fmt_print('[1m')
        
    @property
    def success(self):
        self.fmt_print('[92m')

    @property
    def warning(self):
         self.fmt_print('[93m')

    @property
    def failed(self):
        self.fmt_print('[91m')


@staticmethod
def checkAndGetTwine():
    '''
    - Check if `twine` is installed;
    - Collect `twine`, if it is not;
    '''
    try:
        get_distribution('twine')
        Status('`twine` is installed.').success
    except DistributionNotFound:
        Status('`twine` is not installed.').warning
        check_call([executable, '-m', 'pip', 'install', 'twine'])

def read(*args, **kwargs) -> str:
    with open(join(dirname(__file__), *args), 
         encoding=kwargs.get(*ENCODING)) as fp:
        return fp.read()

def readlines(*args, **kwargs) -> list[str]:
    with open(join(dirname(__file__), *args), 
         encoding=kwargs.get(*ENCODING)) as fp:
        return fp.readlines()

# NOTE :: Get current working directory.

cwd : str = dirname(abspath(__file__))

# NOTE :: Create Essential Files and Installing Essential Packages.

MANIFEST_FNAME : str = 'MANIFEST.ini'
README_FNAME   : str = 'README.md'
LICENSE_FNAME  : str = 'LICENSE.md'
INIT_FNAME     : str = 'snippycode/__init__.py'
VERSION_FNAME  : str = '__version__.py'

if not exists(manifest_path := join(cwd, MANIFEST_FNAME)):
    Status(f'`{MANIFEST_FNAME}` needed to import `{README_FNAME}`.').warning
    with open(manifest_path, 'w', encoding=ENCODING[1]) as manifest:
        print(f'Writing a new `{MANIFEST_FNAME}`.')
        manifest.write(f'include {README_FNAME} {LICENSE_FNAME} requirements.txt\n')
        Status(f'New `{MANIFEST_FNAME}` created.').success

if not exists(init_path := join(cwd, INIT_FNAME)):
    with open(init_path, 'w', encoding=ENCODING[1]) as init:
        init.write(f'')

# NOTE :: Package Medatada.

NAME             : str = 'SnippyCode'
FULLNAME         : str = 'SnippyCode'
VERSION          : str = '0.1.6b0'
DESCRIPTION      : str = 'A command line extension to add snippets directly in Visual Studio Code.'
try:
    long_description : str = read(cwd, README_FNAME)
except FileNotFoundError:
    long_description : str = DESCRIPTION
URL              : str = 'https://github.com/FabioFlorey/SnippyCode'
DOWNLOAD_URL     : str = 'https://pypi.org/project/snippycode/'
KEYWORDS         : list[str] = [
                                'CLI','Command Line Interface',
                                'VSCode', 'Visual Studio Code',
                                'Snippets',
                                'Python 3.9', 'Python 3.10'
                                ]
try:
    license : str = read(cwd, LICENSE_FNAME)
except FileNotFoundError:
    Status(f'`{LICENSE_FNAME}` not found.').failed
    license : str = 'GPL-v3'
    Status(f'Package license will be updated to `{license}`.').warning
    Status(f'Consider adding a `{LICENSE_FNAME}` to your repo!').warning
AUTHOR           : str = 'Fabio Craig Wimmer Florey'
AUTHOR_EMAIL     : str = 'fabioflorey@icloud.com'
MAINTAINER       : str = ''
MAINTAINER_EMAIL : str = ''

# NOTE :: Package Informations.

PYTHON_REQUIRES = '>=3.9.0'

PACKAGES : list[str] = find_packages()
CLASSIFIERS : list[str] = [
                          'Development Status :: 4 - Beta',
                          'Environment :: Console',
                          'Environment :: Win32 (MS Windows)',
                          'Intended Audience :: Developers',
                          'Intended Audience :: End Users/Desktop',
                          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                          'Operating System :: Microsoft :: Windows',
                          'Programming Language :: Python',
                          'Programming Language :: Python :: 3 :: Only',
                          'Topic :: Software Development',
                          'Topic :: Terminals',
                          'Topic :: Utilities'
                         ]

# NOTE :: How to run the script from the CLI.

SNIPPYCODE = 'snippycode.snippyCode'
ENTRYPOINTS : dict[str,
              list[str]] = {
                            'console_scripts': [f'snip={SNIPPYCODE}:write',
                                                f'snip-up={SNIPPYCODE}:update',
                                                f'snip-rm={SNIPPYCODE}:delete',
                                                f'snip-ls={SNIPPYCODE}:list'],
                           }

if __name__ == '__main__':
    setup(
            name=NAME, fullname=FULLNAME, version=VERSION,
            description=DESCRIPTION, long_description=long_description,
            long_description_content_type='text/markdown',
            url=URL, download_url=DOWNLOAD_URL,
            keywords=KEYWORDS, classifiers=CLASSIFIERS,
            license=license,

            author=AUTHOR, author_email=AUTHOR_EMAIL,
            maintainer=MAINTAINER, maintainer_email=MAINTAINER_EMAIL,

            python_requires=PYTHON_REQUIRES,
            packages=PACKAGES,
            entry_points=ENTRYPOINTS
            
         )